/* Solution to the HPCU Bi-Weekly Challenge "Precipitate: Distributed Memory
 * Implementation"
 * Aaron Weeden, Shodor, July 2014
 */

/* LIBS */
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <mpi.h>

/* CONSTANTS */
/** States **/
const int NONE = -1;
const int PRECIPITATING = 0;
const int MOVING = 1;
const int BLOCKED = 2;
const int PRECIPITATE = 3;
const int COLLIDING = 4;

/** Directions **/
const int LEFT = 0;
const int RIGHT = 1;
const int UP = 2;
const int DOWN = 3;

/* INPUT */
int EnvWidth = 20;
int EnvHeight = 20;
int PartWidth = 1;
int PartHeight = 1;
int NParticles = 23;
int Rank;
int NProcs;
int NTimeSteps = 20;
int ChanceCollision = 50;

/* ARRAYS */
int *ParticleStates;
int *MyParticleStates;
int *ParticleDirections;
int *MyParticleDirections;
int *XPositions;
int *MyXPositions;
int *YPositions;
int *MyYPositions;

/* FUNCTION DECLARATIONS */
void readData(int, char**);
void usage();
void randomizePositions();
void pickDirections();
bool tryMove();
void moveParticles();
void checkCollisions();
bool checkPrecipitate(int);
void display(int);

/* MAIN */
int main(int argc, char **argv) {
    int i;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &Rank);
    MPI_Comm_size(MPI_COMM_WORLD, &NProcs);

    srandom(Rank + time(NULL));

    if (Rank == 0) {
        readData(argc, argv);
    }

    MPI_Bcast(&EnvWidth, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&EnvHeight, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&PartWidth, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&PartHeight, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&NParticles, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&NTimeSteps, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&ChanceCollision, 1, MPI_INT, 0, MPI_COMM_WORLD);

    if (NParticles % NProcs != 0) {
        if (Rank == 0) {
            fprintf(stderr, "ERROR: For simplicity's sake, this program only works if the number of particles (specified with -p) is divisible by the number of processes (specified with -np).\n");
        }
        MPI_Finalize();
        exit(EXIT_FAILURE);
    }

    ParticleStates = (int*) malloc(NParticles * sizeof(int));
    MyParticleStates = (int*) malloc(NParticles/NProcs * sizeof(int));
    ParticleDirections = (int*) malloc(NParticles * sizeof(int));
    MyParticleDirections = (int*) malloc(NParticles/NProcs * sizeof(int));
    XPositions = (int*) malloc(NParticles * sizeof(int));
    MyXPositions = (int*) malloc(NParticles/NProcs * sizeof(int));
    YPositions = (int*) malloc(NParticles * sizeof(int));
    MyYPositions = (int*) malloc(NParticles/NProcs * sizeof(int));

    randomizePositions();

    MPI_Allgather(MyXPositions, NParticles/NProcs, MPI_INT,
            XPositions, NParticles/NProcs, MPI_INT, MPI_COMM_WORLD);
    MPI_Allgather(MyYPositions, NParticles/NProcs, MPI_INT,
            YPositions, NParticles/NProcs, MPI_INT, MPI_COMM_WORLD);
    MPI_Allgather(MyParticleStates, NParticles/NProcs, MPI_INT,
            ParticleStates, NParticles/NProcs, MPI_INT, MPI_COMM_WORLD);

    if (Rank == 0) {
        display(0);
    }
    for (i = 0; i < NTimeSteps; i++) {
        pickDirections();

        MPI_Allgather(MyParticleDirections, NParticles/NProcs, MPI_INT,
                ParticleDirections, NParticles/NProcs, MPI_INT, MPI_COMM_WORLD);
        MPI_Allgather(MyParticleStates, NParticles/NProcs, MPI_INT,
                ParticleStates, NParticles/NProcs, MPI_INT, MPI_COMM_WORLD);

        moveParticles();

        MPI_Allgather(MyXPositions, NParticles/NProcs, MPI_INT,
                XPositions, NParticles/NProcs, MPI_INT, MPI_COMM_WORLD);
        MPI_Allgather(MyYPositions, NParticles/NProcs, MPI_INT,
                YPositions, NParticles/NProcs, MPI_INT, MPI_COMM_WORLD);
        MPI_Allgather(MyParticleStates, NParticles/NProcs, MPI_INT,
                ParticleStates, NParticles/NProcs, MPI_INT, MPI_COMM_WORLD);

        if (Rank == 0) {
            checkCollisions();
        }

        MPI_Bcast(ParticleDirections, NParticles, MPI_INT, 0, MPI_COMM_WORLD);
        MPI_Bcast(ParticleStates, NParticles, MPI_INT, 0, MPI_COMM_WORLD);

        if (Rank == 0) {
            display(i+1);
        }
    }

    free(MyYPositions);
    free(YPositions);
    free(MyXPositions);
    free(XPositions);
    free(MyParticleDirections);
    free(ParticleDirections);
    free(MyParticleStates);
    free(ParticleStates);

    MPI_Finalize();

    return 0;
}

/* FUNCTION DEFINITIONS */
void readData(int argc, char **argv) {
    int ch;

    while ((ch = getopt(argc, argv, "w:h:W:H:p:t:c:")) != -1) {
        switch (ch) {
            case 'w':
                EnvWidth = atoi(optarg);
                break;
            case 'h':
                EnvHeight = atoi(optarg);
                break;
            case 'W':
                PartWidth = atoi(optarg);
                break;
            case 'H':
                PartHeight = atoi(optarg);
                break;
            case 'p':
                NParticles = atoi(optarg);
                break;
            case 't':
                NTimeSteps = atoi(optarg);
                break;
            case 'c':
                ChanceCollision = atoi(optarg);
                break;
            case '?':
            default:
                usage(argv);
                exit(EXIT_FAILURE);
        }
    }
}

void usage(char **argv) {
    fprintf(stderr, "USAGE: %s [OPTIONS]\n", argv[0]);
    fprintf(stderr, "  OPTIONS:\n");
    fprintf(stderr, "    -w environment width\n");
    fprintf(stderr, "    -h environment height\n");
    fprintf(stderr, "    -W particle width\n");
    fprintf(stderr, "    -H particle height\n");
    fprintf(stderr, "    -p number of particles\n");
    fprintf(stderr, "    -t number of time steps\n");
    fprintf(stderr, "    -c chance of collision (integer between 0 (inclusive) and 100 (exclusive))\n");
}

void randomizePositions() {
    int i;

    for (i = 0; i < NParticles/NProcs; i++) {
        MyParticleStates[i] = NONE;
        MyXPositions[i] = (EnvWidth - PartWidth) * random() / RAND_MAX;
        MyYPositions[i] = (EnvHeight - PartHeight) * random() / RAND_MAX;
    }
}

void pickDirections() {
    int i;
    bool retval;
    int dir;

    for (i = 0; i < NParticles/NProcs; i++) {
        if (MyParticleStates[i] != COLLIDING &&
                MyParticleStates[i] != PRECIPITATING) {
            retval = false;
            dir = 4 * random() / RAND_MAX;
            switch (dir) {
                case 0:
                    retval = tryMove(i, LEFT);
                    break;
                case 1:
                    retval = tryMove(i, RIGHT);
                    break;
                case 2:
                    retval = tryMove(i, UP);
                    break;
                case 3:
                    retval = tryMove(i, DOWN);
                    break;
            }
            if (retval) {
                MyParticleDirections[i] = dir;
                MyParticleStates[i] = MOVING;
            } else {
                MyParticleStates[i] = BLOCKED;
            }
        }
    }
}

bool tryMove(int i, int dir) {
    int j;

    if (dir == LEFT && MyXPositions[i] == 0) return false;
    else if (dir == RIGHT &&
            MyXPositions[i] == EnvWidth - PartWidth) return false;
    else if (dir == UP && MyYPositions[i] == 0) return false;
    else if (dir == DOWN &&
            MyYPositions[i] == EnvHeight - PartHeight) return false;

    for (j = 0; j < NParticles; j++) {
        if (i+Rank*NParticles/NProcs != j) {
            if (dir == LEFT && XPositions[j] == MyXPositions[i] - 1 &&
                    YPositions[j] == MyYPositions[i]) return false;
            else if (dir == RIGHT && XPositions[j] == MyXPositions[i] + 1 &&
                    YPositions[j] == MyYPositions[i]) return false;
            else if (dir == UP && YPositions[j] == MyYPositions[i] - 1 &&
                    XPositions[j] == MyXPositions[i]) return false;
            else if (dir == DOWN && YPositions[j] == MyYPositions[i] + 1 &&
                    XPositions[j] == MyXPositions[i]) return false;
        }
    }

    return true;
}

void moveParticles() {
    int i;

    for (i = 0; i < NParticles/NProcs; i++) {
        if (MyParticleStates[i] == MOVING) {
            if (MyParticleDirections[i] == LEFT)       MyXPositions[i]--;
            else if (MyParticleDirections[i] == RIGHT) MyXPositions[i]++;
            else if (MyParticleDirections[i] == UP)    MyYPositions[i]--;
            else if (MyParticleDirections[i] == DOWN)  MyYPositions[i]++;
        } else if (MyParticleStates[i] == COLLIDING) {
            MyParticleStates[i] = PRECIPITATING;
        } else if (MyParticleStates[i] == PRECIPITATING) {
            MyYPositions[i]--;
        } else if (checkPrecipitate(i)) {
            MyParticleStates[i] = PRECIPITATE;
        }
    }
}

void checkCollisions() {
    int i;
    int j;

    for (i = 0; i < NParticles; i++) {
        for (j = 0; j < NParticles; j++) {
            if (YPositions[i] == YPositions[j] &&
                    XPositions[i] == XPositions[j] - 1) {
                if ((100 * random() / RAND_MAX) < ChanceCollision) {
                    ParticleStates[i] = ParticleStates[j] = COLLIDING;
                    ParticleDirections[i] = ParticleDirections[j] = DOWN;
                }
            }
            if (ParticleStates[j] == PRECIPITATING &&
                    YPositions[j] == YPositions[i] - 1 &&
                    XPositions[i] == XPositions[j]) {
                ParticleStates[i] = PRECIPITATING;
                ParticleDirections[i] = DOWN;
            }
        }
    }
}

bool checkPrecipitate(int i) {
    int j;

    if (MyYPositions[i] == EnvHeight - PartHeight) return true;
    else
        for (j = 0; j < NParticles; j++)
            if (i+Rank*NParticles/NProcs != j)
                if (ParticleStates[j] == PRECIPITATE &&
                        YPositions[j] == MyYPositions[i] + 1 &&
                        XPositions[j] == MyXPositions[i]) return true;

    return false;
}

void display(int timeStep) {
    int i;
    int s;

    printf("Time step: %d\n", timeStep);
    for (i = 0; i < NParticles; i++) {
        printf("particle %d: (%d,%d) ", i, XPositions[i], YPositions[i]);
        s = ParticleStates[i];
        if (s == PRECIPITATING) printf("precipitating\n");
        else if (s == MOVING) printf("moving\n");
        else if (s == BLOCKED) printf("blocked\n");
        else if (s == PRECIPITATE) printf("precipitate\n");
        else if (s == COLLIDING) printf("colliding\n");
        else if (s == NONE) printf("none\n");
    }
}
